/*
 * $Id: check_icmp.c,v 1.14 2005/03/28 23:19:50 exon Exp $
 *
 * Author: Andreas Ericsson <ae@op5.se>
 *
 * License: GNU GPL 2.0 or any later version.
 *
 * Relevant RFC's: 792 (ICMP), 791 (IP)
 *
 * This program was modeled somewhat after the check_icmp program,
 * which was in turn a hack of fping (www.fping.org) but has been
 * completely rewritten since to generate higher precision rta values,
 * and support several different modes as well as setting ttl to control.
 * redundant routes. The only remainders of fping is currently a few
 * function names.
 *
 * Kudos to;
 *   Joe Rhett
 *   Wickus Botha
 *   Aaron Carr
 *   Harper Mann
 *   Phil Costelloe
 *   Carsten Schmitz
 * for lending me system resources, testing and feedback
 *
 */

#include <sys/time.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#include <signal.h>

/** sometimes undefined system macros (quite a few, actually) **/
#ifndef MAXTTL
# define MAXTTL	255
#endif
#ifndef INADDR_NONE
# define INADDR_NONE 0xffffffU
#endif
#ifndef INADDR_ANY
# define INADDR_ANY 0x000000
#endif

#ifndef SOL_IP
#define SOL_IP 0
#endif

/* we bundle these in one #ifndef, since they're all from BSD
 * Put individual #ifndef's around those that bother you */
#ifndef ICMP_UNREACH_NET_UNKNOWN
# define ICMP_UNREACH_NET_UNKNOWN 6
# define ICMP_UNREACH_HOST_UNKNOWN 7
# define ICMP_UNREACH_ISOLATED 8
# define ICMP_UNREACH_NET_PROHIB 9
# define ICMP_UNREACH_HOST_PROHIB 10
# define ICMP_UNREACH_TOSNET 11
# define ICMP_UNREACH_TOSHOST 12
#endif
/* tru64 has the ones above, but not these */
#ifndef ICMP_UNREACH_FILTER_PROHIB
# define ICMP_UNREACH_FILTER_PROHIB 13
# define ICMP_UNREACH_HOST_PRECEDENCE 14
# define ICMP_UNREACH_PRECEDENCE_CUTOFF 15
#endif


/** typedefs and such **/
enum states {
	STATE_OK = 0,
	STATE_WARNING,
	STATE_CRITICAL,
	STATE_UNKNOWN,
	STATE_DEPENDENT,
	STATE_OOB
};

typedef unsigned short range_t;  /* type for get_range() -- unimplemented */

typedef struct rta_host {
	unsigned short id;           /* id in **table, and icmp pkts */
	char *name;                  /* arg used for adding this host */
	char *msg;                   /* icmp error message, if any */
	struct sockaddr_in saddr_in; /* the address of this host */
	struct in_addr error_addr;   /* stores address of error replies */
	unsigned long int time_waited; /* total time waited, in usecs */
	unsigned int icmp_sent, icmp_recv, icmp_lost; /* counters */
	unsigned char icmp_type, icmp_code; /* type and code from errors */
	unsigned short flags;        /* control/status flags */
	double rta;                  /* measured RTA */
	unsigned char pl;            /* measured packet loss */
	struct rta_host *next;       /* linked list */
} rta_host;

#define FLAG_LOST_CAUSE 0x01  /* decidedly dead target. */

/* threshold structure. all values are maximum allowed, exclusive */
typedef struct threshold {
	unsigned char pl;    /* max allowed packet loss in percent */
	unsigned int rta;  /* roundtrip time average, microseconds */
} threshold;

/* the data structure */
typedef struct icmp_ping_data {
	struct timeval stime;	/* timestamp (saved in protocol struct as well) */
	unsigned short ping_id;
} icmp_ping_data;

/* the different modes of this program are as follows:
 * MODE_RTA: send all packets no matter what (mimic check_icmp and check_ping)
 * MODE_HOSTCHECK: Return immediately upon any sign of life
 *                 In addition, sends packets to ALL addresses assigned
 *                 to this host (as returned by gethostbyname() or
 *                 gethostbyaddr() and expects one host only to be checked at
 *                 a time.  Therefore, any packet response what so ever will
 *                 count as a sign of life, even when received outside
 *                 crit.rta limit. Do not misspell any additional IP's.
 * MODE_ALL:  Requires packets from ALL requested IP to return OK (default).
 * MODE_ICMP: implement something similar to check_icmp (MODE_RTA without
 *            tcp and udp args does this)
 */
#define MODE_RTA 0
#define MODE_HOSTCHECK 1
#define MODE_ALL 2
#define MODE_ICMP 3

/* the different ping types we can do
 * TODO: investigate ARP ping as well */
#define HAVE_ICMP 1
#define HAVE_UDP 2
#define HAVE_TCP 4
#define HAVE_ARP 8

#define MIN_PING_DATA_SIZE sizeof(struct icmp_ping_data)
#define MAX_IP_PKT_SIZE 65536	/* (theoretical) max IP packet size */
#define IP_HDR_SIZE 20
#define MAX_PING_DATA (MAX_IP_PKT_SIZE - IP_HDR_SIZE - ICMP_MINLEN)
#define DEFAULT_PING_DATA_SIZE (MIN_PING_DATA_SIZE + 44)
#define DEFAULT_TTL 64

#define MIN_PKT_INTERVAL   25000 /* min pkt_interval, microsecs */
#define NOFORCE_WAIT 0
#define FORCE_WAIT 1

/** prototypes **/
static void usage(unsigned char, char *, ...);
static u_int get_timevar(const char *);
static u_int get_timevaldiff(struct timeval *, struct timeval *);
static int reap_replies(int, u_int, int);
void u_sleep(u_int u_sec);
static u_int wait_for_reply(int, u_int);
static int recvfrom_wto(int, char *, unsigned int, struct sockaddr *, u_int *);
static int send_icmp_ping(int, struct rta_host *);
static int get_threshold(char *str, threshold *th);
static void run_checks(void);
static int add_target(char *);
static int add_target_ip(char *, struct in_addr *);
static int handle_random_icmp(struct icmp *, struct sockaddr_in *);
static unsigned short icmp_checksum(unsigned short *, int);
static void finish(int);
static void crash(const char *, ...);

/** external **/
extern int optind, opterr, optopt;
/* extern int h_errno; */
extern char *optarg;
extern char **environ;

/** global variables **/
static char *progname;
static struct rta_host **table, *cursor, *list;
static threshold crit = {80, 500000}, warn = {40, 200000};
static int mode, protocols, sockets, debug = 0, timeout = 10;
static unsigned short icmp_pkt_size, icmp_data_size = DEFAULT_PING_DATA_SIZE;
static unsigned int icmp_sent = 0, icmp_recv = 0, icmp_lost = 0;
#define icmp_pkts_en_route (icmp_sent - (icmp_recv + icmp_lost))
static unsigned int targets_down = 0, targets = 0, packets = 0;
#define targets_alive (targets - targets_down)
static unsigned int retry_interval, pkt_interval, target_interval;
static int icmp_sock, tcp_sock, udp_sock, status = STATE_OK;
static pid_t pid;
static struct timezone tz;
static struct timeval prog_start;
static unsigned long int max_completion_time = 0;
static unsigned char ttl = 64;	/* outgoing ttl */
static unsigned int warn_down = 1, crit_down = 1; /* host down threshold values */
float pkt_backoff_factor = 1.5;
float target_backoff_factor = 1.5;
/* compatible hstrerror-like functionality */
const char *txt_resolv_error[] = {
	"Resolver error 0 (no error)",
	"Unknown host",
	"Host name lookup failure",
	"Unknown server error",
	"No address associated with name",
	"Unknown resolver error",
};

/** code start **/
static void
crash(const char *fmt, ...)
{
	va_list ap;

	printf("%s: ", progname);

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);

	if(errno) printf(": %s", strerror(errno));
	puts("");

	exit(3);
}


static char *
icmp_error_msg(unsigned char icmp_type, unsigned char icmp_code)
{
	char *msg = "unreachable";

	if(debug > 4) printf("icmp_error_msg(%u, %u) called\n",
						 icmp_type, icmp_code);
	switch(icmp_type) {
	case ICMP_UNREACH:
		switch(icmp_code) {
		case ICMP_UNREACH_NET: msg = "Net unreachable"; break;
		case ICMP_UNREACH_HOST:	msg = "Host unreachable"; break;
		case ICMP_UNREACH_PROTOCOL: msg = "Protocol unreachable (firewall?)"; break;
		case ICMP_UNREACH_PORT: msg = "Port unreachable (firewall?)"; break;
		case ICMP_UNREACH_NEEDFRAG: msg = "Fragmentation needed"; break;
		case ICMP_UNREACH_SRCFAIL: msg = "Source route failed"; break;
		case ICMP_UNREACH_ISOLATED: msg = "Source host isolated"; break;
		case ICMP_UNREACH_NET_UNKNOWN: msg = "Unknown network"; break;
		case ICMP_UNREACH_HOST_UNKNOWN: msg = "Unknown host"; break;
		case ICMP_UNREACH_NET_PROHIB: msg = "Network administratively prohibited (firewall?)"; break;
		case ICMP_UNREACH_HOST_PROHIB: msg = "Host administratively prohibited (firewall?)"; break;
		case ICMP_UNREACH_TOSNET: msg = "Bad TOS for network (firewall?)"; break;
		case ICMP_UNREACH_TOSHOST: msg = "Bad TOS for host (firewall?)"; break;
		case ICMP_UNREACH_FILTER_PROHIB: msg = "Prohibited by filter (firewall)"; break;
		case ICMP_UNREACH_HOST_PRECEDENCE: msg = "Host precedence violation"; break;
		case ICMP_UNREACH_PRECEDENCE_CUTOFF: msg = "Precedence cutoff"; break;
		default: msg = "Invalid code"; break;
		}
		break;

	case ICMP_TIMXCEED:
		/* really 'out of reach', or non-existant host behind a router serving
		 * two different subnets */
		switch(icmp_code) {
		case ICMP_TIMXCEED_INTRANS: msg = "Time to live exceeded in transit"; break;
		case ICMP_TIMXCEED_REASS: msg = "Fragment reassembly time exceeded"; break;
		default: msg = "Invalid code"; break;
		}
		break;

	case ICMP_ECHOREPLY: msg = "Echo reply"; break;
	case ICMP_ECHO: msg = "Echo request"; break;
	case ICMP_SOURCEQUENCH: msg = "Transmitting too fast"; break;
	case ICMP_REDIRECT: msg = "Redirect (change route)"; break;
	case ICMP_PARAMPROB: msg = "Bad IP header (required option absent)"; break;

		/* the following aren't error messages, so ignore */
	case ICMP_TSTAMP: msg = "Timestamp request"; break;
	case ICMP_TSTAMPREPLY: msg = "Timestamp reply"; break;
	case ICMP_IREQ: msg = "Information request"; break;
	case ICMP_IREQREPLY: msg = "Information reply"; break;
	case ICMP_MASKREQ: msg = "Address mask request"; break;
	case ICMP_MASKREPLY: msg = "Address mask reply"; break;
	default: msg = "Invalid ICMP type (kernel IP stack corrupt?)"; break;
	}

	return msg;
}

static int
handle_random_icmp(struct icmp *p, struct sockaddr_in *addr)
{
	struct icmp *sent_icmp = NULL;
	struct rta_host *host = NULL;
	unsigned char *ptr;

	if(p->icmp_type == ICMP_ECHO && p->icmp_id == pid) {
		if(debug > 2) printf("echo request from us to us (pinging localhost)\n");
		return 0;
	}

	ptr = (unsigned char *)p;
	if(debug) printf("handle_random_icmp(%p, %p)\n", (void *)p, (void *)addr);

	/* only handle a few types, since others can't possibly be replies to
	 * us in a sane network (if it is anyway, it will be counted as lost
	 * at summary time, but not as quickly as a proper response */
	/* TIMXCEED can be an unreach from a router with multiple IP's which
	 * serves two different subnets on the same interface and a dead host
	 * on one net is pinged from the other. The router will respond to
	 * itself and thus set TTL=0 so as to not loop forever.  Even when
	 * TIMXCEED actually sends a proper icmp response we will have passed
	 * too many hops to have a hope of reaching it later, in which case it
	 * indicates overconfidence in the network, poor routing or both. */
	if(debug > 2) printf("random ICMP response type: %u,  code: %u\n",
						 p->icmp_type, p->icmp_code);
	if(p->icmp_type != ICMP_UNREACH && p->icmp_type != ICMP_TIMXCEED &&
	   p->icmp_type != ICMP_SOURCEQUENCH && p->icmp_type != ICMP_PARAMPROB)
	{
		if(debug > 1) printf("Unsafe to handle unknown type %u, returning\n",
							 p->icmp_type);
		return 0;
	}

	/* might be for us. At least it holds the original package (according
	 * to RFC 792). If it isn't, just ignore it */
	sent_icmp = (struct icmp *)(ptr + 28);
	if(sent_icmp->icmp_type != ICMP_ECHO || sent_icmp->icmp_id != pid ||
	   sent_icmp->icmp_seq >= targets)
	{
		if(debug) printf("not a response to a packet we sent, or target failed to duplicate request\n");
		return 0;
	}

	/* it is indeed a response for us */
	host = table[sent_icmp->icmp_seq];
	if(debug) {
		printf("Received \"%s\" from %s for ICMP ECHO sent to %s.\n",
			   icmp_error_msg(p->icmp_type, p->icmp_code),
			   inet_ntoa(addr->sin_addr), host->name);
	}

	icmp_lost++;
	host->icmp_lost++;
	/* don't spend time on lost hosts any more */
	if(host->flags & FLAG_LOST_CAUSE) return 0;

	/* source quench means we're sending too fast, so increase the
	 * interval and mark this packet lost */
	if(p->icmp_type == ICMP_SOURCEQUENCH) {
		pkt_interval *= pkt_backoff_factor;
		target_interval *= target_backoff_factor;
	}
	else {
		targets_down++;
		host->flags |= FLAG_LOST_CAUSE;
	}

	host->icmp_type = p->icmp_type;
	host->icmp_code = p->icmp_code;
	host->error_addr.s_addr = addr->sin_addr.s_addr;

	return 0;
}

int
main(int argc, char **argv)
{
	int i;
	char *ptr;
	long int arg;
	int icmp_sockerrno, udp_sockerrno, tcp_sockerrno;
	int result;
	struct rta_host *host;
	
	/* we only need to be setsuid when we get the sockets, so do
	 * that before pointer magic (esp. on network data) */
	icmp_sockerrno = udp_sockerrno = tcp_sockerrno = sockets = 0;

	if((icmp_sock = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP)) != -1)
		sockets |= HAVE_ICMP;
	else icmp_sockerrno = errno;

	/* if((udp_sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) != -1) */
	/* 	sockets |= HAVE_UDP; */
	/* else udp_sockerrno = errno; */

	/* if((tcp_sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) != -1) */
	/* 	sockets |= HAVE_TCP; */
	/* else tcp_sockerrno = errno; */

	/* now drop privileges (no effect if not setsuid or geteuid() == 0) */
	setuid(getuid());

	/* POSIXLY_CORRECT might break things, so unset it (the portable way) */
	environ = NULL;

	/* use the pid to mark packets as ours */
	pid = getpid();

	/* get calling name the old-fashioned way for portability instead
	 * of relying on the glibc-ism __progname */
	ptr = strrchr(argv[0], '/');
	if(ptr) progname = &ptr[1];
	else progname = argv[0];

	/* now set defaults. Use progname to set them initially (allows for
	 * superfast check_host program when target host is up */
	cursor = list = NULL;
	table = NULL;

	mode = MODE_RTA;
	crit.rta = 500000;
	crit.pl = 80;
	warn.rta = 200000;
	warn.pl = 40;
	protocols = HAVE_ICMP | HAVE_UDP | HAVE_TCP;
	pkt_interval = 80000;  /* 80 msec packet interval by default */
	packets = 5;

	if(!strcmp(progname, "check_icmp") || !strcmp(progname, "check_ping")) {
		mode = MODE_ICMP;
		protocols = HAVE_ICMP;
	}
	else if(!strcmp(progname, "check_host")) {
		mode = MODE_HOSTCHECK;
		pkt_interval = 200000;
		packets = 5;
		crit.rta = warn.rta = 1000000;
		crit.pl = warn.pl = 100;
	}
	else if(!strcmp(progname, "check_icmp_multi")) {
		mode = MODE_ALL;
		target_interval = 0;
		pkt_interval = 50000;
		packets = 5;
	}

	/* parse the arguments */
	for(i = 1; i < argc; i++) {
		/* the leading '-' of the optstring makes non-argument options
		 * return with arg 1, meaning it's */
		while((arg = getopt(argc, argv, "-vhVw:c:n:p:t:H:i:b:I:l:")) != EOF) {
			switch(arg) {
			case 'v':
				debug++;
				break;
			case 'b':
				/* silently ignored for now */
				break;
			case 'i':
				pkt_interval = get_timevar(optarg);
				break;
			case 'I':
				target_interval = get_timevar(optarg);
				break;
			case 'w':
				get_threshold(optarg, &warn);
				break;
			case 'c':
				get_threshold(optarg, &crit);
				break;
			case 'n':
			case 'p':
				packets = strtoul(optarg, NULL, 0);
				break;
			case 't':
				timeout = strtoul(optarg, NULL, 0);
				if(!timeout) timeout = 10;
				break;
			case 'H': case 1: /* 1 is default for non-option arguments */
				if(debug > 2) printf("arg: %ld, optarg: %s\n", arg, optarg);
				add_target(optarg);
				break;
			case 'l':
				ttl = (unsigned char)strtoul(optarg, NULL, 0);
				break;
			case 'd': /* implement later, for cluster checks */
				warn_down = (unsigned char)strtoul(optarg, &ptr, 0);
				if(ptr) {
					crit_down = (unsigned char)strtoul(ptr + 1, NULL, 0);
				}
				break;
			case 'h': case 'V': default:
				usage(arg, NULL);
				break;
			}
		}
	}

	if(debug > 3) printf("First non-option argument is %d (%s)\n", optind, argv[optind]);
	argv = &argv[optind];
	while(*argv) {
		add_target(*argv);
		argv++;
	}
	if(!targets) usage('h', "No hosts to check");

	if(getuid() && pkt_interval < MIN_PKT_INTERVAL)
		pkt_interval = MIN_PKT_INTERVAL;

	if(debug > 1)
		printf("uid: %d, pid: %d\n", getuid(), pid);

	if(!sockets) {
		if(icmp_sock == -1) {
			errno = icmp_sockerrno;
			crash("Failed to obtain ICMP socket");
			return -1;
		}
		/* if(udp_sock == -1) { */
		/* 	errno = icmp_sockerrno; */
		/* 	crash("Failed to obtain UDP socket"); */
		/* 	return -1; */
		/* } */
		/* if(tcp_sock == -1) { */
		/* 	errno = icmp_sockerrno; */
		/* 	crash("Failed to obtain TCP socker"); */
		/* 	return -1; */
		/* } */
	}
	if(!ttl) ttl = 64;

	if(icmp_sock) {
		errno = 0;
		result = setsockopt(icmp_sock, SOL_IP, IP_TTL, &ttl, sizeof(ttl));
		if(debug > 2) {
			printf("setsockopt(icmp_sock, SOL_IP, IP_TTL, %u) = %d (%s)\n",
				   ttl, result, strerror(errno));
			if(result != -1) printf("outgoing ttl set to %u\n", ttl);
		}
	}

	/* stupid users should be able to give whatever thresholds they want
	 * (nothing will break if they do), but some anal plugin maintainer
	 * will probably add some printf() thing here later, so it might be
	 * best to at least show them where to do it. ;) */
	if(warn.pl > crit.pl) warn.pl = crit.pl;
	if(warn.rta > crit.rta) warn.rta = crit.rta;
	if(warn_down > crit_down) crit_down = warn_down;

	signal(SIGINT, finish);
	signal(SIGHUP, finish);
	signal(SIGTERM, finish);
	signal(SIGALRM, finish);
	if(debug > 1) printf("Setting alarm timeout to %u seconds\n", timeout);
	alarm(timeout);

	/* make sure we don't wait any longer than necessary */
	gettimeofday(&prog_start, &tz);
	/* 100msec higher than absolute minimum to allow for cpu time */
	max_completion_time =
		((targets * packets * pkt_interval) + (targets * target_interval)) +
		(targets * packets * crit.rta) + 100000;

	if(debug > 1) {
		printf("packets: %u, targets: %u\n"
			   "target_interval: %0.3f, pkt_interval %0.3f, crit.rta: %0.3f\n"
			   "max_completion_time: %0.3f\n",
			   packets, targets,
			   (float)target_interval / 1000, (float)pkt_interval / 1000,
			   (float)crit.rta / 1000,
			   (float)max_completion_time / 1000);
	}

	if(debug) {
		if(max_completion_time > (u_int)timeout * 1000000) {
			printf("max_completion_time: %lu  timeout: %u\n",
				   max_completion_time, timeout);
			printf("Timout must be at lest %lu\n",
				   max_completion_time / 1000000 + 1);
		}
	}

	icmp_pkt_size = icmp_data_size + ICMP_MINLEN;
	if(debug > 2) printf("icmp_pkt_size = %u\n", icmp_pkt_size);
	if(icmp_pkt_size < sizeof(struct icmp) + sizeof(struct icmp_ping_data)) {
		icmp_pkt_size = sizeof(struct icmp) + sizeof(struct icmp_ping_data);
	}
	if(debug > 2) printf("icmp_pkt_size = %u\n", icmp_pkt_size);

	if(debug) {
		printf("crit = {%u, %u%%}, warn = {%u, %u%%}\n",
			   crit.rta, crit.pl, warn.rta, warn.pl);
		printf("pkt_interval: %u  target_interval: %u  retry_interval: %u\n",
			   pkt_interval, target_interval, retry_interval);
		printf("icmp_pkt_size: %u  timeout: %u\n",
			   icmp_pkt_size, timeout);
	}

	if(packets > 20) {
		errno = 0;
		crash("packets is > 20 (%d)", packets);
	}

	host = list;
	table = malloc(sizeof(struct rta_host **) * (argc - 1));
	i = 0;
	while(host) {
		host->id = i;
		table[i] = host;
		host = host->next;
		i++;
	}

	run_checks();

	errno = 0;
	finish(0);

	return(0);
}

static void
run_checks()
{
	u_int i, t, result;
	u_int final_wait, time_passed;

	/* send the packets */
	for(t = 0; t < targets; t++) {
		for(i = 0; i < packets; i++) {

			/* don't send useless packets */
			if(table[t]->flags & FLAG_LOST_CAUSE) {
				if(debug) printf("%s is a lost cause. not sending any more\n",
								 table[t]->name);
				continue;
			}

			/* we're still in the game, so send next packet */
			(void)send_icmp_ping(icmp_sock, table[t]);

			/* pause for a while so we don't flood the network */
			result = reap_replies(icmp_sock, pkt_interval, FORCE_WAIT);
		}

		/* wait for all incoming packets from all hosts */
		if(target_interval > pkt_interval)
			result = reap_replies(icmp_sock, target_interval - pkt_interval, FORCE_WAIT);
	}

	if(icmp_pkts_en_route && targets_alive) {
		time_passed = get_timevaldiff(NULL, NULL);
		final_wait = max_completion_time - time_passed;

		if(debug) {
			printf("time_passed: %u  final_wait: %u  max_completion_time: %lu\n",
				   time_passed, final_wait, max_completion_time);
		}
		if(time_passed > max_completion_time) {
			if(debug) printf("Time passed. Finishing up\n");
			finish(0);
		}

		/* catch the packets that might come in within the timeframe, but
		 * haven't yet */
		if(debug) printf("Waiting for %u micro-seconds (%0.3f msecs)\n",
						 final_wait, (float)final_wait / 1000);
		result = reap_replies(icmp_sock, final_wait, NOFORCE_WAIT);
	}
}

/*
 * select() is posix, so we expect it to be around
 */
void
u_sleep(u_int u_sec)
{
	int nfound;
	struct timeval to;
	fd_set readset, writeset;

	if(debug > 3) printf("sleeping for %u microseconds\n", u_sec);
	if(!u_sec) return;

	to.tv_sec = u_sec / 1000000;
	to.tv_usec = u_sec % 1000000;
	FD_ZERO(&writeset);
	FD_ZERO(&readset);
	nfound = select(0, &readset, &writeset, NULL, &to);
	if(nfound < 0)
		crash("select() in u_sleep");

	return;
}								/* u_sleep() */

static int
reap_replies(int sock, u_int t, int wait_forced)
{
	/* wrap up if all targets are declared dead */
	if(debug > 3) printf("reap_replies(%d, %u, %d) called\n", sock, t, wait_forced);

	if(!targets_alive ||
	   get_timevaldiff(&prog_start, NULL) >= max_completion_time)
	{
		finish(0);
	}

	/* listen til timeout while there are packets en route (rhymes!) ;) */
	while(t && icmp_pkts_en_route) {
		t = wait_for_reply(sock, t);

		/* the dead won't sing to us */
		if(!targets_alive) finish(0);
	}

	if(icmp_recv == (targets_alive * (int)packets)) finish(0);

	if(!t || !wait_forced) return 0;

	u_sleep(t);

	return 0;
}

/* response structure:
 * ip header   : 20 bytes
 * icmp header : 28 bytes
 * icmp echo reply : the rest
 */
static u_int
wait_for_reply(int sock, u_int t)
{
	int n, hlen;
	static char buf[4096];
	struct sockaddr_in resp_addr;
	struct ip *ip;
	struct icmp *icp;
	struct rta_host *host;
	struct icmp_ping_data *data;
	struct timeval wait_start, now;
	u_int tdiff, timo = t;

	gettimeofday(&wait_start, &tz);

	n = recvfrom_wto(sock, buf, sizeof(buf), (struct sockaddr *)&resp_addr, &timo);
	if(!n) {	/* timeout */
		if(debug > 2) printf("recvfrom_wto() timed out during a %u usecs wait\n", t);
		return 0; /* no time left, so return just that */
	}

	gettimeofday(&now, &tz);
	tdiff = get_timevaldiff(&wait_start, &now);
	/* make sure we can return time left */
	if(t > tdiff) t -= tdiff;
	else t = 0;

	ip = (struct ip *)buf;
	if(debug > 1) {
		printf("received %u bytes from %s\n",
			   ntohs(ip->ip_len), inet_ntoa(resp_addr.sin_addr));
		if(ntohs(ip->ip_len) == 16384)
			printf("Possible FreeBSD 5.3 recvfrom(2) weirdness.\n");
		if(ntohs(ip->ip_len) != n)
			printf("recvfrom() (%d) and ntohs(ip->ip_len) (%u) disagree (broken IP stack or voodoo?!)\n",
				   n, ntohs(ip->ip_len));
	}

/* obsolete, or at least not globally applicable */
/* #if defined( __alpha__ ) && __STDC__ && !defined( __GLIBC__ ) */
		/* alpha headers are decidedly broken. Using an ansi compiler,
		 * they provide ip_vhl instead of ip_hl and ip_v, so we mask
		 * off the bottom 4 bits */
/* 		hlen = (ip->ip_vhl & 0x0f) << 2; */
/* #else */
	hlen = ip->ip_hl << 2;
/* #endif */

	if(n < (hlen + ICMP_MINLEN)) {
		crash("received packet too short for ICMP (%d bytes, expected %d) from %s\n",
			  n, hlen + icmp_pkt_size, inet_ntoa(resp_addr.sin_addr));
	}
	else if(debug > 2) {
		printf("ip header size: %u, packet size: %u (expected %u, %u)\n",
			   hlen, ntohs(ip->ip_len) - hlen,
			   sizeof(struct ip), (u_int)icmp_pkt_size);
	}

	/* check the response */
	icp = (struct icmp *)(buf + hlen);
	if(debug)
		printf("inbound packet (type: %d, code: %d, %s) is:\n",
			   icp->icmp_type, icp->icmp_code,
			   icmp_error_msg(icp->icmp_type, icp->icmp_code));

	/* some systems have 32 bit pid_t, so mask off bottom 16 bits of icmp_id */
	if(icp->icmp_id != (pid & 0xffff) || icp->icmp_type != ICMP_ECHOREPLY || icp->icmp_seq >= targets) {
		if(debug) {
			if(icp->icmp_id != pid)
				printf("\tnot marked by me (icp->icmp_id: %d, pid: %d)\n",
					   icp->icmp_id, pid);
			if(icp->icmp_type != ICMP_ECHOREPLY) printf("\tnot an echo reply\n");
			if(icp->icmp_seq >= targets) printf("\ttagged with a too high seq number\n");
		}
		handle_random_icmp(icp, &resp_addr);
		return t;
	}
	if(debug) puts("\ta valid ICMP_ECHOREPLY to a packet we sent");

	/* this is indeed a valid response */
	data = (struct icmp_ping_data *)(icp->icmp_data);
	host = table[icp->icmp_seq];
	tdiff = get_timevaldiff(&data->stime, &now);

	host->time_waited += tdiff;
	host->icmp_recv++;
	icmp_recv++;

	if(debug) {
		printf("%0.3f ms rtt from %s, outgoing ttl: %u, incoming ttl: %u\n",
			   (float)tdiff / 1000, inet_ntoa(resp_addr.sin_addr),
			   ttl, ip->ip_ttl);
	}

	/* if we're in hostcheck mode, exit with limited printouts */
	if(mode == MODE_HOSTCHECK) {
		printf("OK - %s responds to ICMP. Packet %u, rtt %0.3fms|"
			   "pkt=%u;;0;%u rta=%0.3f;%0.3f;%0.3f;;\n",
			   host->name, icmp_recv, (float)tdiff / 1000,
			   icmp_recv, packets, (float)tdiff / 1000,
			   (float)warn.rta / 1000, (float)crit.rta / 1000);
		exit(STATE_OK);
	}

	return t;
}

/* the ping functions */
static int
send_icmp_ping(int sock, struct rta_host *host)
{
	static char *buf = NULL; /* re-use so we prevent leaks */
	long int len;
	struct icmp *icp;
	struct icmp_ping_data *data;
	struct timeval tv;
	struct sockaddr *addr;

	
	if(sock == -1) {
		errno = 0;
		crash("Attempt to send on bogus socket");
		return -1;
	}
	addr = (struct sockaddr *)&host->saddr_in;

	if(!buf) {
		buf = (char *)malloc(icmp_pkt_size + sizeof(struct ip));
		if(!buf) {
			crash("send_icmp_ping(): failed to malloc %d bytes for send buffer",
				  icmp_pkt_size);
			return -1;	/* might be reached if we're in debug mode */
		}
	}
	memset(buf, 0, icmp_pkt_size + sizeof(struct ip));

	icp = (struct icmp *)buf;
	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_id = (pid & 0xffff); /* 32 bit pid_t is possible, so mask 16 */
	icp->icmp_seq = host->id;
	data = (struct icmp_ping_data *)icp->icmp_data;
	data->ping_id = host->icmp_sent;

	if((gettimeofday(&tv, &tz)) == -1) return -1;
	memcpy(&data->stime, &tv, sizeof(struct timeval));
	icp->icmp_cksum = icmp_checksum((u_short *)icp, icmp_pkt_size);

	len = sendto(sock, buf, icmp_pkt_size, 0, (struct sockaddr *)addr,
				 sizeof(struct sockaddr));

	if(len < 0 || (unsigned int)len != icmp_pkt_size) {
		if(debug) printf("Failed to send ping to %s\n",
						 inet_ntoa(host->saddr_in.sin_addr));
		return -1;
	}

	/* increment the counters */
	icmp_sent++;
	host->icmp_sent++;

	return 0;
}

static int
recvfrom_wto(int sock, char *buf, unsigned int len, struct sockaddr *saddr,
			 u_int *timo)
{
	u_int slen;
	int n, result;
	struct timeval to, then, now;
	fd_set rd, wr;

	if(!*timo) {
		if(debug) printf("*timo is not\n");
		return 0;
	}

	to.tv_sec = *timo / 1000000;
	to.tv_usec = (*timo - (to.tv_sec * 1000000));

	FD_ZERO(&rd);
	FD_ZERO(&wr);
	FD_SET(sock, &rd);
	errno = 0;
	gettimeofday(&then, &tz);
	n = select(sock + 1, &rd, &wr, NULL, &to);
	if(n < 0) crash("select() in recvfrom_wto");
	gettimeofday(&now, &tz);
	*timo = get_timevaldiff(&then, &now);

	if(!n) return 0;				/* timeout */

	slen = sizeof(struct sockaddr);

	result = recvfrom(sock, buf, len, 0, saddr, &slen);
	if(debug > 3) printf("recvfrom() returned %d\n", result);
	if(result < 0) crash("recvfrom in recvfrom_wto");

	return result;
}

static void
finish(int sig)
{
	u_int i = 0;
	unsigned char pl;
	double rta;
	struct rta_host *host;
	char *status_string[] =
	{"OK", "WARNING", "CRITICAL", "UNKNOWN", "DEPENDENT"};

	alarm(0);
	if(debug > 2) printf("finish(%d) called\n", sig);

	if(icmp_sock != -1) close(icmp_sock);
	if(udp_sock != -1) close(udp_sock);
	if(tcp_sock != -1) close(tcp_sock);

	if(debug) {
		printf("icmp_sent: %u  icmp_recv: %u  icmp_lost: %u\n",
			   icmp_sent, icmp_recv, icmp_lost);
		printf("targets: %u  targets_alive: %u\n", targets, targets_alive);
	}

	/* if we're in hostcheck mode we only end up here in case we didn't get
	 * any packets, so finish up quickly */
	host = list;
	if(mode == MODE_HOSTCHECK) {
		printf("%s is DOWN - ", list->name);
		/* print all icmp errors we encountered (if any) */
		while(host) {
			if(!host->icmp_type) continue;
			printf("%s @ %s; ",
				   icmp_error_msg(host->icmp_type, host->icmp_code),
				   inet_ntoa(host->error_addr));
			host = host->next;
		}

		printf("rta: nan, lost 100%%|pkt=%d;%d;%d;%d;%d pl=100%%;95;100;;\n",
			   packets + 1, packets, packets, packets, packets);

		exit(STATE_CRITICAL);
	}

	/* iterate thrice to calculate values, give output, and print perfparse */
	while(host) {
		if(!host->icmp_recv) {
			/* rta 0 is ofcourse not entirely correct, but will still show up
			 * conspicuosly as missing entries in perfparse and cacti */
			pl = 100;
			rta = 0;
			status = STATE_CRITICAL;
			/* up the down counter if not already counted */
			if(!(host->flags & FLAG_LOST_CAUSE) && targets_alive) targets_down++;
		}
		else {
			pl = ((host->icmp_sent - host->icmp_recv) * 100) / host->icmp_sent;
			rta = (double)host->time_waited / host->icmp_recv;
		}
		host->pl = pl;
		host->rta = rta;
		if(!status && (pl >= warn.pl || rta >= warn.rta)) status = STATE_WARNING;
		if(pl >= crit.pl || rta >= crit.rta) status = STATE_CRITICAL;

		host = host->next;
	}
	/* this is inevitable */
	if(!targets_alive) status = STATE_CRITICAL;
	printf("%s - ", status_string[status]);

	host = list;
	while(host) {
		if(debug) puts("");
		if(i) {
			if(i < targets) printf(" :: ");
			else printf("\n");
		}
		i++;
		if(!host->icmp_recv) {
			status = STATE_CRITICAL;
			if(host->flags & FLAG_LOST_CAUSE) {
				printf("%s: %s @ %s. rta nan, lost %d%%",
					   host->name,
					   icmp_error_msg(host->icmp_type, host->icmp_code),
					   inet_ntoa(host->error_addr),
					   100);
			}
			else { /* not marked as lost cause, so we have no flags for it */
				printf("%s: rta nan, lost 100%%", host->name);
			}
		}
		else {	/* !icmp_recv */
			printf("%s: rta %0.3fms, lost %u%%",
				   host->name, host->rta / 1000, host->pl);
		}

		host = host->next;
	}

	/* iterate once more for pretty perfparse output */
	printf("|");
	i = 0;
	host = list;
	while(host) {
		if(debug) puts("");
		printf("%srta=%0.3fms;%0.3f;%0.3f;0; %spl=%u%%;%u;%u;; ",
			   (targets > 1) ? host->name : "",
			   host->rta / 1000, (float)warn.rta / 1000, (float)crit.rta / 1000,
			   (targets > 1) ? host->name : "",
			   host->pl, warn.pl, crit.pl);

		host = host->next;
	}

	/* finish with an empty line */
	puts("");
	if(debug) printf("targets: %u, targets_alive: %u\n",
					 targets, targets_alive);

	exit(status);
}

static u_int
get_timevaldiff(struct timeval *early, struct timeval *later)
{
	u_int ret;
	struct timeval now;

	if(!later) {
		gettimeofday(&now, &tz);
		later = &now;
	}
	if(!early) early = &prog_start;

	/* if early > later we return 0 so as to indicate a timeout */
	if(early->tv_sec > early->tv_sec ||
	   (early->tv_sec == later->tv_sec && early->tv_usec > later->tv_usec))
	{
		return 0;
	}

	ret = (later->tv_sec - early->tv_sec) * 1000000;
	ret += later->tv_usec - early->tv_usec;

	return ret;
}

static int
add_target_ip(char *arg, struct in_addr *in)
{
	struct rta_host *host;

	/* disregard obviously stupid addresses */
	if(in->s_addr == INADDR_NONE || in->s_addr == INADDR_ANY) {
		if(debug > 2) printf("Discarding stupid address %s\n", inet_ntoa(*in));
		return -1;
	}

	/* no point in adding two identical IP's, so don't. ;) */
	host = list;
	while(host) {
		if(host->saddr_in.sin_addr.s_addr == in->s_addr) {
			if(debug) printf("Identical IP already exists. Not adding %s\n", arg);
			return -1;
		}
		host = host->next;
	}

	/* add the fresh ip */
	host = malloc(sizeof(struct rta_host));
	if(!host) {
		crash("add_target_ip(%s, %s): malloc(%d) failed",
			  arg, inet_ntoa(*in), sizeof(struct rta_host));
	}
	memset(host, 0, sizeof(struct rta_host));

	/* set the values. use calling name for output */
	host->name = strdup(arg);

	/* fill out the sockaddr_in struct */
	host->saddr_in.sin_family = AF_INET;
	host->saddr_in.sin_addr.s_addr = in->s_addr;

	if(!list) list = cursor = host;
	else cursor->next = host;

	cursor = host;
	targets++;

	return 0;
}

/* wrapper for add_target_ip */
static int
add_target(char *arg)
{
	int i;
	struct hostent *he;
	struct in_addr *in, ip;

	/* don't resolve if we don't have to and don't add all ip's if we
	 * were given a specific one, even if we're in hostcheck mode */
	if(debug > 2) printf("Adding target %s\n", arg);
	ip.s_addr = inet_addr(arg);
	if(ip.s_addr != INADDR_NONE) return add_target_ip(arg, &ip);

	/* not an IP, so resolve */
	errno = 0;
	he = gethostbyname(arg);
	if(!he && h_errno == TRY_AGAIN) {
		u_sleep(500000);
		he = gethostbyname(arg);
	}

	if(!he) {
		if(h_errno > 5) h_errno = 5;
		usage(0, "Failed to resolve %s: %s", arg, txt_resolv_error[h_errno]);
	}

	/* possibly add all the IP's as targets */
	for(i = 0; he->h_addr_list[i]; i++) {
		in = (struct in_addr *)he->h_addr_list[i];
		add_target_ip(arg, in);

		/* this is silly, but it works */
		if(mode == MODE_HOSTCHECK || mode == MODE_ALL) {
			if(debug) printf("mode: %d\n", mode);
			continue;
		}
		break;
	}

	return 0;
}

/*
 * u = micro
 * m = milli
 * s = seconds
 * return value is in microseconds
 */
static u_int
get_timevar(const char *str)
{
	char p, u, *ptr;
	unsigned int len;
	u_int i, d;	            /* integer and decimal, respectively */
	u_int factor = 1000;    /* default to milliseconds */

	if(!str) return 0;
	len = strlen(str);
	if(!len) return 0;

	/* unit might be given as ms|m (millisec),
	 * us|u (microsec) or just plain s, for seconds */
	u = p = '\0';
	u = str[len - 1];
	if(len >= 2 && !isdigit((int)str[len - 2])) p = str[len - 2];
	if(p && u == 's') u = p;
	else if(!p) p = u;
	if(debug > 3) printf("evaluating %s, u: %c, p: %c\n", str, u, p);

	if(u == 'u') factor = 1;            /* microseconds */
	else if(u == 'm') factor = 1000;	/* milliseconds */
	else if(u == 's') factor = 1000000;	/* seconds */
	if(debug > 3) printf("factor is %u\n", factor);

	i = strtoul(str, &ptr, 0);
	if(!ptr || *ptr != '.' || strlen(ptr) < 2 || factor == 1)
		return i * factor;

	/* time specified in usecs can't have decimal points, so ignore them */
	if(factor == 1) return i;

	d = strtoul(ptr + 1, NULL, 0);

	/* d is decimal, so get rid of excess digits */
	while(d >= factor) d /= 10;

	/* the last parenthesis avoids floating point exceptions. */
	return ((i * factor) + (d * (factor / 10)));
}

/* not too good at checking errors, but it'll do (main() should barfe on -1) */
static int
get_threshold(char *str, threshold *th)
{
	char *p = NULL, i = 0;

	if(!str || !strlen(str) || !th) return -1;

	/* pointer magic slims code by 10 lines. i is bof-stop on stupid libc's */
	p = &str[strlen(str) - 1];
	while(p != &str[1]) {
		if(*p == '%') *p = '\0';
		else if(*p == ',' && i) {
			*p = '\0';	/* reset it so get_timevar(str) works nicely later */
			th->pl = (unsigned char)strtoul(p+1, NULL, 0);
			break;
		}
		i = 1;
		p--;
	}
	th->rta = get_timevar(str);

	if(!th->rta) return -1;

	if(th->rta > MAXTTL * 1000000) th->rta = MAXTTL * 1000000;
	if(th->pl > 100) th->pl = 100;

	return 0;
}

unsigned short
icmp_checksum(unsigned short *p, int n)
{
	register unsigned short cksum;
	register long sum = 0;

	while(n > 1) {
		sum += *p++;
		n -= 2;
	}

	/* mop up the occasional odd byte */
	if(n == 1) sum += (unsigned char)*p;

	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	cksum = ~sum;				/* ones-complement, trunc to 16 bits */

	return cksum;
}

/* make core plugin developers happy (silly, really) */
static void
usage(unsigned char arg, char *fmt, ...)
{
	if(fmt) {
		va_list ap;

		printf("%s: ", progname);

		va_start(ap, fmt);
		vprintf(fmt, ap);
		va_end(ap);
		puts("\n");
	}

	if(arg == 'V') {
		printf("$Id: check_icmp.c,v 1.14 2005/03/28 23:19:50 exon Exp $\n");
		exit(STATE_UNKNOWN);
	}

	printf("Usage: %s [options] [-H] host1 host2 hostn\n\n", progname);

	if(arg != 'h') {
		printf("%s -h  for details\n", progname);
		exit(3);
	}

	printf("Where options are any combination of:\n"
		   " * -H | --host      specify a target\n"
		   " * -w | --warn      warning threshold (currently %0.3fms,%u%%)\n"
		   " * -c | --crit      critical threshold (currently %0.3fms,%u%%)\n"
		   " * -n | --packets   number of packets to send (currently %u)\n"
		   " * -i | --interval  max packet interval (currently %0.3fms)\n"
		   " * -I | --hostint   max target interval (currently %0.3fms)\n",
		   (float)warn.rta / 1000, warn.pl, (float)crit.rta / 1000, crit.pl,
		   packets,
		   (float)pkt_interval / 1000, (float)target_interval / 1000);

	printf(" * -l | --ttl       outgoing TTL (currently %u, not supported on all platforms)\n"
		   " * -t | --timeout   timeout value (seconds, currently  %u)\n"
		   " * -b | --bytes     icmp packet size (currenly ignored)\n"
		   "   -v | --verbose   verbosity++ (4 is sort of max)\n"
		   "   -h | --help      this cruft\n",
		   ttl, timeout);

	puts("\nThreshold format for -w and -c is 200.25,60% for 200.25 msec RTA and 60%\n"
		 "packet loss.  All threshold values match inclusively.\n"
		 "You can specify different RTA factors using the standardized abbreviations\n"
		 "us (microseconds), ms (milliseconds, default) or just plain s for seconds.\n");
	puts("Long options are currently unsupported.\n");
	puts("Options marked with * require an argument\n");
	puts("If this program is invoked as check_host (with a symlink, preferrably), it will exit with status\n"
		 "OK upon the first properly received ICMP_ECHOREPLY, making it ideal\n"
		 "for hostchecks (less network load and much, much faster on OK)\n");

	exit(3);
}
